﻿using System;
using System.IO;
using System.Collections;
using System.Collections.Generic;
using UnityEngine;
using UnityEditor;

public class VarjoHDRPPatch : MonoBehaviour
{
#if HDRP_7_OR_NEWER && !(HDRP_7_5_1_OR_NEWER && !HDRP_8_OR_NEWER) && !HDRP_10_OR_NEWER
    [MenuItem("Edit/Render Pipeline/Fix HDRP bugs affecting Varjo XR Plugin", false, 2000)]
    static void Patch()
    {
        if (!Application.isPlaying)
        {
            string sourceDirectory = "Packages/com.unity.render-pipelines.high-definition";
            string directoryName = new DirectoryInfo(Path.GetFullPath(sourceDirectory)).Name;
            string destinationDirectory = string.Format("Assets/../Packages/{0}", directoryName);

            int knownBugs = 0;
            int bugsFixed = 0;

            bool HDRPBug = false;
            bool XRSystemBug = false;

#if HDRP_7_4_OR_NEWER || (HDRP_8_OR_NEWER && !HDRP_10_OR_NEWER)
            HDRPBug = true;
            knownBugs++;
#endif

#if !HDRP_7_4_3_OR_NEWER || (HDRP_8_OR_NEWER && !HDRP_10_OR_NEWER)
            XRSystemBug = true;
            knownBugs++;
#endif

            if (knownBugs == 0)
            {
                Debug.Log("No known bugs to patch.");
                return;
            }

            if (EditorUtility.DisplayDialog("Patch HDRP bugs affecting Varjo XR plugin?", string.Format("This operation moves HDRP package from Package Cache to <Project>/Packages and modifies {0} files." +
                "\n\nDo you want to continue?", knownBugs), "Continue", "Cancel"))
            {
                if (!Directory.Exists(destinationDirectory))
                {
                    try
                    {
                        FileUtil.MoveFileOrDirectory(sourceDirectory, destinationDirectory);
                        Debug.Log(string.Format("Moved HDRP package from {0} to {1} to enable modifications.", sourceDirectory, destinationDirectory));
                    }
                    catch (Exception e)
                    {
                        Debug.LogError(string.Format("Couldn't move HDRP package from {0} to {1} to enable modifications. Operation failed with error: {0}", sourceDirectory, destinationDirectory, e.Message));
                        return;
                    }
                }

                if (HDRPBug)
                {
                    // Apply fix https://github.com/Unity-Technologies/Graphics/pull/1133
                    if (PatchHDRP(destinationDirectory)) bugsFixed++;
                }

                if (XRSystemBug)
                {
                    // Apply fix https://github.com/Unity-Technologies/Graphics/pull/1269
                    if (PatchXRSystem(destinationDirectory)) bugsFixed++;
                }

                Debug.Log(string.Format("{0} of the {1} known bugs fixed. All modifications are preceded by comment // Varjo patch", bugsFixed, knownBugs));
            }
        }
        else
        {
            Debug.LogWarning("Can't patch HDRP bugs in Play Mode. Exit Play Mode and try again.");
        }

    }
    #endif

    static bool PatchHDRP(string packagePath)
    {
        string HDRPSourcePath = string.Format("{0}/Runtime/RenderPipeline/HDRenderPipeline.cs", packagePath);
        string[] HDRPSource = File.ReadAllLines(HDRPSourcePath);
        string[] HDRPPatch = {
            "                                    // Varjo patch",
            "                                    // Applies fix from https://github.com/Unity-Technologies/Graphics/pull/1133",
            "                                    m_SkyManager.UpdateCurrentSkySettings(hdCamera);",
            "                                    // End of Varjo patch"
        };

        foreach(string line in HDRPSource)
        {
            if (line.Contains("Varjo patch")) {
                Debug.Log(string.Format("{0} has already been patched.", HDRPSourcePath));
                return true;
            }
        }

        string[] patchedHDRPSource = new string[HDRPSource.Length + HDRPPatch.Length];

        bool insideBlock = false;
        bool patched = false;
        int k = 0;
        for (int i = 0; i < HDRPSource.Length; i++)
        {
            if (!patched && !insideBlock && HDRPSource[i].Contains("if (camera == req.hdCamera.camera && req.hdCamera.xr.cullingPassId == xrPass.cullingPassId)"))
            {
                insideBlock = true;
            }
            else if (insideBlock)
            {
                if (HDRPSource[i].Contains("m_SkyManager.UpdateCurrentSkySettings(hdCamera);"))
                {
                    Debug.Log(string.Format("{0} already contains a fix for the bug.", HDRPSourcePath));
                    return true;
                } else if (HDRPSource[i].Contains("}"))
                {
                    for (k = 0; k < HDRPPatch.Length; k++)
                    {
                        patchedHDRPSource[i + k] = HDRPPatch[k];
                    }
                    insideBlock = false;
                    patched = true;
                }
            }
            patchedHDRPSource[i + k] = HDRPSource[i];
        }
        File.WriteAllText(HDRPSourcePath, string.Empty);
        File.WriteAllLines(HDRPSourcePath, patchedHDRPSource);

        Debug.Log(string.Format("{0} patched succesfully!", HDRPSourcePath));
        return true;
    }

    static bool PatchXRSystem(string packagePath)
    {
        string XRSystemSourcePath = string.Format("{0}/Runtime/RenderPipeline/XR/XRSystem.cs", packagePath);
        string[] XRSystemSource = File.ReadAllLines(XRSystemSourcePath);
        string[] XRSystemPatch = {
            "                    // Varjo patch",
            "                    // Applies fix from https://github.com/Unity-Technologies/Graphics/pull/1133",
            "                    if (display != null)",
            "                    {",
            "                        display.zNear = camera.nearClipPlane;",
            "                        display.zFar = camera.farClipPlane;",
            "                    }",
            "                    // End of Varjo patch"
        };

        foreach (string line in XRSystemSource)
        {
            if (line.Contains("Varjo patch"))
            {
                Debug.Log(string.Format("{0} has already been patched.", XRSystemSourcePath));
                return true;
            }
        }

        string[] patchedXRSystemSource = new string[XRSystemSource.Length + XRSystemPatch.Length];

        bool patched = false;
        int k = 0;
        for (int i = 0; i < XRSystemSource.Length; i++)
        {
            if (XRSystemSource[i].Contains("display.zNear = camera.nearClipPlane;"))
            {
                Debug.Log(string.Format("{0} already contains a fix for the bug.", XRSystemSourcePath));
                return true;
            }
            else if (!patched && XRSystemSource[i].Contains("CreateLayoutFromXrSdk(camera, singlePassAllowed);"))
            {
                for (k = 0; k < XRSystemPatch.Length; k++)
                {
                    patchedXRSystemSource[i + k] = XRSystemPatch[k];
                }
                patched = true;
            }
            patchedXRSystemSource[i + k] = XRSystemSource[i];
        }
        File.WriteAllText(XRSystemSourcePath, string.Empty);
        File.WriteAllLines(XRSystemSourcePath, patchedXRSystemSource);

        Debug.Log(string.Format("{0} patched succesfully!", XRSystemSourcePath));
        return true;
    }

}
